"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ImportValue(CloudFormationLintRule):
    """Check if a Output is done of another output"""

    id = "W6001"
    shortdesc = "Check Outputs using ImportValue"
    description = "Check if the Output value is set using ImportValue, so creating an Output of an Output"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["outputs", "importvalue"]

    def match(self, cfn):
        matches = []

        # Get all import values
        importvalue_trees = cfn.search_deep_keys("Fn::ImportValue")

        # Filter out the Outputs
        importvalue_trees = [x for x in importvalue_trees if x[0] == "Outputs"]

        # Check if the importvalue is set on the Value
        for importvalue_tree in importvalue_trees:
            # Skip invalid configuration, let other rules handle that
            if len(importvalue_tree) < 4:
                continue

            if importvalue_tree[2] == "Value":
                # ImportValue can be used within other intrinic function, exclude those
                if importvalue_tree[3] == "Fn::ImportValue":
                    message = "The value of output ({0}) is imported from another output ({1})"
                    matches.append(
                        RuleMatch(
                            importvalue_tree,
                            message.format(importvalue_tree[1], importvalue_tree[-1]),
                        )
                    )

        return matches



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from typing import List, Optional, Sequence, Union

from cfnlint.rules import Match, RulesCollection
from cfnlint.template import Template

LOGGER = logging.getLogger(__name__)


class Runner:
    """Run all the rules"""

    def __init__(
        self,
        rules: RulesCollection,
        filename: Optional[str],
        template: str,
        regions: Sequence[str],
        verbosity=0,
        mandatory_rules: Union[Sequence[str], None] = None,
    ):
        self.rules = rules
        self.filename = filename
        self.verbosity = verbosity
        self.mandatory_rules = mandatory_rules or []
        self.cfn = Template(filename, template, regions)

    def transform(self):
        """Transform logic"""
        matches = self.cfn.transform()
        return matches

    def run(self) -> List[Match]:
        """Run rules"""
        LOGGER.info("Run scan of template %s", self.filename)
        matches = []
        if self.cfn.template is not None:
            matches.extend(self.rules.run(self.filename, self.cfn))
        return self.check_metadata_directives(matches)

    def check_metadata_directives(self, matches: Sequence[Match]) -> List[Match]:
        # uniq the list of incidents and filter out exceptions from the template
        directives = self.cfn.get_directives()
        return_matches: List[Match] = []
        for match in matches:
            if not any(match == u for u in return_matches):
                if match.rule.id not in directives:
                    return_matches.append(match)
                else:
                    for mandatory_rule in self.mandatory_rules:
                        if match.rule.id.startswith(mandatory_rule):
                            return_matches.append(match)
                            break
                    else:
                        path = getattr(match, "path", None)
                        if path:
                            if len(path) >= 2:
                                if path[0] != "Resources":
                                    return_matches.append(match)
                                    continue
                                if path[1] not in directives[match.rule.id]:
                                    return_matches.append(match)
                            else:
                                return_matches.append(match)

        return return_matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class SnapStartEnabled(CloudFormationLintRule):
    """Check if the SnapStart is enabled for certain java runtimes"""

    id = "I2530"
    shortdesc = "Validate that SnapStart is configured for >= Java11 runtimes"
    description = (
        "SnapStart is a no-cost feature that can increase performance up to 10x. "
        "Enable SnapStart for Java 11 and greater runtimes"
    )
    source_url = "https://docs.aws.amazon.com/lambda/latest/dg/snapstart.html"
    tags = ["resources", "lambda"]

    def __init__(self):
        super().__init__()
        self.resource_property_types.append("AWS::Lambda::Function")

    def validate(self, runtime, path, region, regions):
        if not isinstance(runtime, str):
            return []

        if region not in regions:
            return []

        if not (runtime.startswith("java")) or runtime in ["java8.al2", "java8"]:
            return []

        return [
            RuleMatch(
                path,
                f"When using {runtime} configure SnapStart",
                rule=self,
            )
        ]



"""
Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
from typing import Any, Dict, List, Mapping, Tuple, Union

from cfnlint.conditions._utils import get_hash

LOGGER = logging.getLogger(__name__)
REF_REGION = get_hash({"Ref": "AWS::Region"})


class EqualParameter:
    hash: str = ""

    def __init__(self, value: dict):
        self._value = value
        self.hash: str = get_hash(value)

    def __eq__(self, __o: Any):
        if isinstance(__o, str):
            return self.hash == __o
        return self.hash == __o.hash


class Equal:
    hash: str
    _left: Union[EqualParameter, str]
    _right: Union[EqualParameter, str]
    _is_static: Union[bool, None]
    _is_region: Tuple[bool, str]

    def __init__(self, equal: List[Union[str, dict]]) -> None:
        self._is_static = None
        if isinstance(equal, list) and len(equal) == 2:
            # sort to keep consistancy from random ordering
            # pylint: disable=unnecessary-lambda
            equal_s = sorted(equal, key=lambda value: json.dumps(value))

            self._left = self._init_parameter(equal_s[0])
            self._right = self._init_parameter(equal_s[1])

            self.hash = get_hash([self._left, self._right])
            if isinstance(self._left, str) and isinstance(self._right, str):
                self._is_static = self._left == self._right
            elif isinstance(self._left, EqualParameter) and isinstance(
                self._right, EqualParameter
            ):
                if self._left == self._right:
                    self._is_static = True

            self._is_region = (False, "")
            if isinstance(self._left, EqualParameter):
                if self._left.hash == REF_REGION and isinstance(self._right, str):
                    self._is_region = (True, self._right)
            if isinstance(self._right, EqualParameter):
                if self._right.hash == REF_REGION and isinstance(self._left, str):
                    self._is_region = (True, self._left)

            return
        raise ValueError("Equals has to be a list of two values")

    def _init_parameter(
        self, parameter: Union[Dict, str]
    ) -> Union[EqualParameter, str]:
        if isinstance(parameter, dict):
            return EqualParameter(parameter)
        return str(parameter)

    @property
    def is_static(self) -> Union[bool, None]:
        """Returns a boolean value if the result is always True or False or None if
            it isn't a static boolean

        Args: None

        Returns:
            Union[bool, None]: None if the equals can be True or False or True/False if
            the equals will always return the same result
        """
        return self._is_static

    @property
    def parameters(self) -> List[EqualParameter]:
        """Returns a List of the EqualParameter that make up the Condition

        Args: None

        Returns:
            List[Equal]: A list of the left and right equal parameters if they are
                         of type EqualParameter
        """
        params = []
        if isinstance(self._left, EqualParameter):
            params.append(self._left)
        if isinstance(self._right, EqualParameter):
            params.append(self._right)
        return params

    @property
    def is_region(self) -> Tuple[bool, str]:
        """Returns a Tuple if the condition is comparing a region to a string

        Args: None

        Returns:
            Tuple[bool, str]: Tuple where the boolean is True if the condition
              is using Ref: AWS::Region and the second element is for the region
              being compared
        """
        return self._is_region

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    def test(self, scenarios: Mapping[str, str]) -> bool:
        """Do an equals based on the provided scenario"""
        if self._is_static in [True, False]:
            return self._is_static
        for scenario, value in scenarios.items():
            if isinstance(self._left, EqualParameter):
                if scenario == self._left:
                    return value == self._right
            if isinstance(self._right, EqualParameter):
                if scenario == self._right:
                    return value == self._left

        raise ValueError("An appropriate scenario was not found")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class ServerlessTransform(CloudFormationLintRule):
    """Check if Serverless Resources exist without the Serverless Transform"""

    id = "E3038"
    shortdesc = "Check if Serverless Resources have Serverless Transform"
    description = "Check that a template with Serverless Resources also includes the Serverless Transform"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint"
    tags = ["resources", "transform"]

    def match(self, cfn):
        matches = []

        transforms = cfn.template.get("Transform", [])
        if not isinstance(transforms, list):
            transforms = [transforms]
        has_serverless_transform = any(
            transform == "AWS::Serverless-2016-10-31" for transform in transforms
        )
        if has_serverless_transform:
            return matches

        for resource_name, resource_values in cfn.get_resources().items():
            resource_type = resource_values["Type"]
            if isinstance(resource_type, str):
                if resource_type.startswith("AWS::Serverless::"):
                    message = (
                        "Serverless Transform required for Type {0} for resource {1}"
                    )
                    matches.append(
                        RuleMatch(
                            ["Transform"], message.format(resource_type, resource_name)
                        )
                    )
                    break
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import LIMITS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class LimitValue(CloudFormationLintRule):
    """Check if maximum Parameter value size limit is exceeded"""

    id = "E2012"
    shortdesc = "Parameter value limit not exceeded"
    description = "Check if the size of Parameter values in the template is less than the upper limit"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["parameters", "limits"]

    def match(self, cfn):
        matches = []

        value_limit = LIMITS["Parameters"]["value"]

        # There are no real "Values" in the template, check the "meta" information
        # (Default, AllowedValue and MaxLength) against the limit
        for paramname, paramvalue in cfn.get_parameters_valid().items():
            # Check Default value
            default_value = paramvalue.get("Default")

            if isinstance(default_value, (str)):
                if len(default_value) > value_limit:
                    path = ["Parameters", paramname, "Default"]
                    message = "The length of parameter default value ({0}) exceeds the limit ({1})"
                    matches.append(
                        RuleMatch(path, message.format(len(default_value), value_limit))
                    )

            # Check MaxLength parameters
            max_length = paramvalue.get("MaxLength", 0)

            if isinstance(max_length, (str)):
                try:
                    max_length = int(max_length)
                except ValueError:
                    # Configuration errors are not the responsibility of this rule
                    max_length = 0

            if isinstance(max_length, int):
                if max_length > value_limit:
                    path = ["Parameters", paramname, "MaxLength"]
                    message = "The MaxLength of parameter ({0}) exceeds the limit ({1})"
                    matches.append(
                        RuleMatch(path, message.format(max_length, value_limit))
                    )

            # Check AllowedValues
            allowed_values = paramvalue.get("AllowedValues", [])

            for allowed_value in allowed_values:
                if isinstance(allowed_value, (str)):
                    if len(allowed_value) > value_limit:
                        path = ["Parameters", paramname, "AllowedValues"]
                        message = "The length of parameter allowed value ({0}) exceeds the limit ({1})"
                        matches.append(
                            RuleMatch(
                                path, message.format(len(allowed_value), value_limit)
                            )
                        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Description(CloudFormationLintRule):
    """Check if Outputs Descriptions are only string values"""

    id = "E6005"
    shortdesc = "Outputs descriptions can only be strings"
    description = "Outputs descriptions can only be strings"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html"
    tags = ["outputs"]

    def match(self, cfn):
        matches = []

        outputs = cfn.get_outputs_valid()
        if outputs:
            for output_name, output_value in outputs.items():
                description = output_value.get("Description")
                if description:
                    if not isinstance(description, str):
                        message = "Output Description can only be a string"
                        matches.append(
                            RuleMatch(["Outputs", output_name, "Description"], message)
                        )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json

from cfnlint.helpers import RESOURCE_SPECS
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class AllowedValue(CloudFormationLintRule):
    """Check if properties have a valid value"""

    id = "E3030"
    shortdesc = "Check if properties have a valid value"
    description = "Check if properties have a valid value in case of an enumator"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#allowedvalue"
    tags = ["resources", "property", "allowed value"]

    def initialize(self, cfn):
        """Initialize the rule"""
        for resource_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "ResourceTypes"
        ):
            self.resource_property_types.append(resource_type_spec)
        for property_type_spec in RESOURCE_SPECS.get(cfn.regions[0]).get(
            "PropertyTypes"
        ):
            self.resource_sub_property_types.append(property_type_spec)

    def check_value(self, value, path, property_name, **kwargs):
        """Check Value"""
        matches = []
        allowed_value_specs = kwargs.get("value_specs", {}).get("AllowedValues", {})

        if allowed_value_specs:
            # Ignore values with dynamic references. Simple check to prevent false-positives
            # See: https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html
            if "{{resolve:" not in str(value):
                # Always compare the allowed value as a string, strict typing is not of concern for this rule
                if str(value) not in allowed_value_specs:
                    message = "You must specify a valid value for {0} ({1}). {2}"
                    description = f"Valid values are {json.dumps(allowed_value_specs)}"
                    matches.append(
                        RuleMatch(
                            path, message.format(property_name, value, description)
                        )
                    )

        return matches

    def check(self, cfn, properties, value_specs, property_specs, path):
        """Check itself"""
        matches = []
        for p_value, p_path in properties.items_safe(path[:]):
            for prop in p_value:
                if prop in value_specs:
                    value = value_specs.get(prop).get("Value", {})
                    if value:
                        value_type = value.get("ValueType", "")
                        property_type = (
                            property_specs.get("Properties").get(prop).get("Type")
                        )
                        value_type_details = (
                            RESOURCE_SPECS.get(cfn.regions[0])
                            .get("ValueTypes")
                            .get(value_type, {})
                        )
                        if value_type_details == "CACHED":
                            value_type_details = (
                                RESOURCE_SPECS.get("us-east-1")
                                .get("ValueTypes")
                                .get(value_type, {})
                            )
                        matches.extend(
                            cfn.check_value(
                                p_value,
                                prop,
                                p_path,
                                check_value=self.check_value,
                                value_specs=value_type_details,
                                cfn=cfn,
                                property_type=property_type,
                                property_name=prop,
                            )
                        )

        return matches

    def match_resource_sub_properties(self, properties, property_type, path, cfn):
        """Match for sub properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("PropertyTypes")
            .get(property_type, {})
            .get("Properties", {})
        )
        property_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("PropertyTypes").get(property_type)
        )
        matches.extend(self.check(cfn, properties, specs, property_specs, path))

        return matches

    def match_resource_properties(self, properties, resource_type, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        specs = (
            RESOURCE_SPECS.get(cfn.regions[0])
            .get("ResourceTypes")
            .get(resource_type, {})
            .get("Properties", {})
        )
        resource_specs = (
            RESOURCE_SPECS.get(cfn.regions[0]).get("ResourceTypes").get(resource_type)
        )
        matches.extend(self.check(cfn, properties, specs, resource_specs, path))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from difflib import SequenceMatcher

import cfnlint.helpers
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Properties(CloudFormationLintRule):
    """Check Base Resource Configuration"""

    id = "E3002"
    shortdesc = "Resource properties are invalid"
    description = "Making sure that resources properties are properly configured"
    source_url = "https://github.com/aws-cloudformation/cfn-python-lint/blob/main/docs/cfn-resource-specification.md#properties"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.cfn = {}
        self.resourcetypes = {}
        self.propertytypes = {}
        self.parameternames = {}
        self.intrinsictypes = {}

    def primitivetypecheck(self, value, primtype, proppath):
        """
        Check primitive types.
        Only check that a primitive type is actual a primitive type:
        - If its JSON let it go
        - If its Conditions check each sub path of the condition
        - If its a object make sure its a valid function and function
        - If its a list raise an error

        """

        matches = []
        supported_functions = [
            "Fn::Base64",
            "Fn::GetAtt",
            "Fn::GetAZs",
            "Fn::ImportValue",
            "Fn::Join",
            "Fn::Split",
            "Fn::FindInMap",
            "Fn::Select",
            "Ref",
            "Fn::If",
            "Fn::Contains",
            "Fn::Sub",
            "Fn::Cidr",
            "Fn::Transform",
            "Fn::Length",
            "Fn::ToJsonString",
        ]
        if (isinstance(value, list) and primtype != "Json") or value == {
            "Ref": "AWS::NotificationARNs"
        }:
            message = f'Property should be of type {primtype} not List at {"/".join(map(str, proppath))}'
            matches.append(RuleMatch(proppath, message))
        if isinstance(value, dict) and primtype == "Json":
            return matches
        if isinstance(value, dict):
            if len(value) == 1:
                for sub_key, sub_value in value.items():
                    if sub_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                        # not erroring on bad Ifs but not need to account for it
                        # so the rule doesn't error out
                        if isinstance(sub_value, list):
                            if len(sub_value) == 3:
                                matches.extend(
                                    self.primitivetypecheck(
                                        sub_value[1], primtype, proppath + ["Fn::If", 1]
                                    )
                                )
                                matches.extend(
                                    self.primitivetypecheck(
                                        sub_value[2], primtype, proppath + ["Fn::If", 2]
                                    )
                                )
                    elif sub_key not in supported_functions:
                        message = f'Property {"/".join(map(str, proppath))} has an illegal function {sub_key}'
                        matches.append(RuleMatch(proppath, message))
            else:
                message = f'Property is an object instead of {primtype} at {"/".join(map(str, proppath))}'
                matches.append(RuleMatch(proppath, message))

        return matches

    def _check_list_for_condition(
        self, text, prop, parenttype, resourcename, propspec, path
    ):
        """Loop for conditions"""
        matches = []
        if len(text) == 3:
            for if_i, if_v in enumerate(text[1:]):
                condition_path = path[:] + [if_i + 1]
                if isinstance(if_v, list):
                    for index, item in enumerate(if_v):
                        arrproppath = condition_path[:]

                        arrproppath.append(index)
                        matches.extend(
                            self.propertycheck(
                                item,
                                propspec["ItemType"],
                                parenttype,
                                resourcename,
                                arrproppath,
                                False,
                            )
                        )
                elif isinstance(if_v, dict):
                    if len(if_v) == 1:
                        for d_k, d_v in if_v.items():
                            if d_k != "Ref" or d_v != "AWS::NoValue":
                                if d_k == "Fn::GetAtt":
                                    resource_name = None
                                    if isinstance(d_v, list):
                                        resource_name = d_v[0]
                                    elif isinstance(d_v, str):
                                        resource_name = d_v.split(".")[0]
                                    if resource_name:
                                        resource_type = (
                                            self.cfn.template.get("Resources", {})
                                            .get(resource_name, {})
                                            .get("Type")
                                        )
                                        if not resource_type.startswith("Custom::"):
                                            message = "Property {0} should be of type List for resource {1} at {2}"
                                            matches.append(
                                                RuleMatch(
                                                    condition_path,
                                                    message.format(
                                                        prop,
                                                        resourcename,
                                                        (
                                                            "/".join(
                                                                str(x)
                                                                for x in condition_path
                                                            )
                                                        ),
                                                    ),
                                                )
                                            )
                                elif d_k == "Fn::If":
                                    matches.extend(
                                        self._check_list_for_condition(
                                            d_v,
                                            prop,
                                            parenttype,
                                            resourcename,
                                            propspec,
                                            condition_path,
                                        )
                                    )
                                else:
                                    message = "Property {0} should be of type List for resource {1} at {2}"
                                    matches.append(
                                        RuleMatch(
                                            condition_path,
                                            message.format(
                                                prop,
                                                resourcename,
                                                (
                                                    "/".join(
                                                        str(x) for x in condition_path
                                                    )
                                                ),
                                            ),
                                        )
                                    )
                    else:
                        message = "Property {0} should be of type List for resource {1} at {2}"
                        matches.append(
                            RuleMatch(
                                condition_path,
                                message.format(
                                    prop,
                                    resourcename,
                                    ("/".join(str(x) for x in condition_path)),
                                ),
                            )
                        )
                else:
                    message = (
                        "Property {0} should be of type List for resource {1} at {2}"
                    )
                    matches.append(
                        RuleMatch(
                            condition_path,
                            message.format(
                                prop,
                                resourcename,
                                ("/".join(str(x) for x in condition_path)),
                            ),
                        )
                    )

        else:
            message = f'Invalid !If condition specified at {"/".join(map(str, path))}'
            matches.append(RuleMatch(path, message))

        return matches

    def check_list_for_condition(
        self, text, prop, parenttype, resourcename, propspec, path
    ):
        """Checks lists that are a dict for conditions"""
        matches = []
        if len(text[prop]) == 1:  # pylint: disable=R1702
            for sub_key, sub_value in text[prop].items():
                if sub_key in cfnlint.helpers.CONDITION_FUNCTIONS:
                    matches.extend(
                        self._check_list_for_condition(
                            sub_value,
                            prop,
                            parenttype,
                            resourcename,
                            propspec,
                            path + [sub_key],
                        )
                    )
                else:
                    # FindInMaps can be lists of objects so skip checking those
                    if sub_key != "Fn::FindInMap":
                        # if its a GetAtt to a custom resource that custom resource
                        # can return a list of objects so skip.
                        if sub_key == "Fn::GetAtt":
                            resource_name = None
                            if isinstance(sub_value, list):
                                resource_name = sub_value[0]
                            elif isinstance(sub_value, str):
                                resource_name = sub_value.split(".")[0]
                            if resource_name:
                                resource_type = (
                                    self.cfn.template.get("Resources", {})
                                    .get(resource_name, {})
                                    .get("Type")
                                )
                                if not (
                                    resource_type
                                    == "AWS::CloudFormation::CustomResource"
                                    or resource_type.startswith("Custom::")
                                ):
                                    message = f'Property is an object instead of List at {"/".join(map(str, path))}'
                                    matches.append(RuleMatch(path, message))
                        elif not (sub_key == "Ref" and sub_value == "AWS::NoValue"):
                            message = f'Property is an object instead of List at {"/".join(map(str, path))}'
                            matches.append(RuleMatch(path, message))
                    else:
                        self.logger.debug(
                            'Too much logic to handle whats actually in the map "%s" so skipping any more validation.',
                            sub_value,
                        )
        else:
            message = (
                f'Property is an object instead of List at {"/".join(map(str, path))}'
            )
            matches.append(RuleMatch(path, message))

        return matches

    def check_exceptions(self, parenttype, proptype, text):
        """
        Checks for exceptions to the spec
        - Start with handling exceptions for templated code.
        """
        templated_exceptions = {
            "AWS::ApiGateway::RestApi": ["S3Location"],
            "AWS::Lambda::Function": ["Code"],
            "AWS::Lambda::LayerVersion": ["Content"],
            "AWS::ElasticBeanstalk::ApplicationVersion": ["SourceBundle"],
            "AWS::StepFunctions::StateMachine": ["S3Location"],
        }

        exceptions = templated_exceptions.get(parenttype, [])
        if proptype in exceptions:
            if isinstance(text, str):
                return True

        return False

    def propertycheck(self, text, proptype, parenttype, resourcename, path, root):
        """Check individual properties"""

        parameternames = self.parameternames
        matches = []
        if root:
            specs = self.resourcetypes
            resourcetype = parenttype
        else:
            specs = self.propertytypes
            resourcetype = str.format("{0}.{1}", parenttype, proptype)
            # Handle tags
            if resourcetype not in specs:
                if proptype in specs:
                    resourcetype = proptype
                else:
                    resourcetype = str.format("{0}.{1}", parenttype, proptype)
            else:
                resourcetype = str.format("{0}.{1}", parenttype, proptype)

        resourcespec = specs[resourcetype].get("Properties", {})
        if not resourcespec:
            if specs[resourcetype].get("Type") == "List":
                if isinstance(text, list):
                    property_type = specs[resourcetype].get("ItemType")
                    for index, item in enumerate(text):
                        matches.extend(
                            self.propertycheck(
                                item,
                                property_type,
                                parenttype,
                                resourcename,
                                path[:] + [index],
                                root,
                            )
                        )

            return matches
        supports_additional_properties = specs[resourcetype].get(
            "AdditionalProperties", False
        )

        if text == "AWS::NoValue":
            return matches
        if not isinstance(text, dict):
            if not self.check_exceptions(parenttype, proptype, text):
                message = f'Expecting an object at {"/".join(map(str, path))}'
                matches.append(RuleMatch(path, message))
            return matches

        # You can put in functions directly in place of objects as long as that is
        # the only thing there (conditions, select) could all (possibly)
        # return objects.  FindInMap cannot directly return an object.
        len_of_text = len(text)

        # pylint: disable=too-many-nested-blocks
        for prop in text:
            proppath = path[:]
            proppath.append(prop)
            if prop not in resourcespec:
                if prop in cfnlint.helpers.CONDITION_FUNCTIONS and len_of_text == 1:
                    cond_values = self.cfn.get_condition_values(text[prop])
                    for cond_value in cond_values:
                        if isinstance(cond_value["Value"], dict):
                            matches.extend(
                                self.propertycheck(
                                    cond_value["Value"],
                                    proptype,
                                    parenttype,
                                    resourcename,
                                    proppath + cond_value["Path"],
                                    root,
                                )
                            )
                        elif isinstance(cond_value["Value"], list):
                            for index, item in enumerate(cond_value["Value"]):
                                matches.extend(
                                    self.propertycheck(
                                        item,
                                        proptype,
                                        parenttype,
                                        resourcename,
                                        proppath + cond_value["Path"] + [index],
                                        root,
                                    )
                                )
                elif text.is_function_returning_object():
                    self.logger.debug(
                        'Ran into function "%s".  Skipping remaining checks', prop
                    )
                elif (
                    len(text) == 1
                    and prop in "Ref"
                    and text.get(prop) == "AWS::NoValue"
                ):
                    pass
                elif len(text) == 1 and prop in "Fn::GetAtt":
                    getatt = text.get(prop)
                    getatt_type = (
                        self.cfn.template.get("Resources", {})
                        .get(getatt[0], {})
                        .get("Type", "")
                    )
                    if (
                        getatt_type == "AWS::CloudFormation::CustomResource"
                        or getatt_type.startswith("Custom::")
                    ):
                        pass
                    else:
                        message = f'GetAtt must refer to a custom resource {"/".join(map(str, proppath))}'
                        matches.append(RuleMatch(proppath, message))
                elif not supports_additional_properties:
                    close_match = False
                    for key in resourcespec.keys():
                        if SequenceMatcher(a=prop, b=key).ratio() > 0.8:
                            message = f'Invalid Property {"/".join(map(str, proppath))}. Did you mean {key}?'
                            matches.append(RuleMatch(proppath, message))
                            close_match = True
                            break
                    if not close_match:
                        message = f'Invalid Property {"/".join(map(str, proppath))}'
                        matches.append(RuleMatch(proppath, message))
            else:
                if "Type" in resourcespec[prop]:
                    if resourcespec[prop]["Type"] == "List":
                        if "PrimitiveItemType" not in resourcespec[prop]:
                            if isinstance(text[prop], list):
                                for index, item in enumerate(text[prop]):
                                    arrproppath = proppath[:]
                                    arrproppath.append(index)
                                    matches.extend(
                                        self.propertycheck(
                                            item,
                                            resourcespec[prop]["ItemType"],
                                            parenttype,
                                            resourcename,
                                            arrproppath,
                                            False,
                                        )
                                    )
                            elif isinstance(text[prop], dict):
                                # A list can be be specific as a Conditional
                                matches.extend(
                                    self.check_list_for_condition(
                                        text,
                                        prop,
                                        parenttype,
                                        resourcename,
                                        resourcespec[prop],
                                        proppath,
                                    )
                                )
                            else:
                                message = "Property {0} should be of type List for resource {1}"
                                matches.append(
                                    RuleMatch(
                                        proppath, message.format(prop, resourcename)
                                    )
                                )
                        else:
                            if isinstance(text[prop], list):
                                primtype = resourcespec[prop]["PrimitiveItemType"]
                                for index, item in enumerate(text[prop]):
                                    arrproppath = proppath[:]
                                    arrproppath.append(index)
                                    matches.extend(
                                        self.primitivetypecheck(
                                            item, primtype, arrproppath
                                        )
                                    )
                            elif isinstance(text[prop], dict):
                                if "Ref" in text[prop]:
                                    ref = text[prop]["Ref"]
                                    if ref == "AWS::NotificationARNs":
                                        continue
                                    if ref in parameternames:
                                        param_type = self.cfn.template["Parameters"][
                                            ref
                                        ]["Type"]
                                        if param_type:
                                            if (
                                                "List<" not in param_type
                                                and "<List" not in param_type
                                                and "<CommaDelimitedList"
                                                not in param_type
                                                and not param_type
                                                == "CommaDelimitedList"
                                            ):
                                                message = (
                                                    "Property {0} should be of type List or Parameter should "
                                                    "be a list for resource {1}"
                                                )
                                                matches.append(
                                                    RuleMatch(
                                                        proppath,
                                                        message.format(
                                                            prop, resourcename
                                                        ),
                                                    )
                                                )
                                    else:
                                        message = "Property {0} should be of type List for resource {1}"
                                        matches.append(
                                            RuleMatch(
                                                proppath,
                                                message.format(prop, resourcename),
                                            )
                                        )
                                else:
                                    if len(text[prop]) == 1:
                                        for k in text[prop].keys():
                                            def_intrinsic_type = (
                                                self.intrinsictypes.get(k, {})
                                            )
                                            if def_intrinsic_type:
                                                if (
                                                    len(
                                                        def_intrinsic_type.get(
                                                            "ReturnTypes"
                                                        )
                                                    )
                                                    == 1
                                                    and def_intrinsic_type.get(
                                                        "ReturnTypes"
                                                    )[0]
                                                    == "Singular"
                                                ):
                                                    message = "Property {0} is using {1} when a List is needed for resource {2}"
                                                    matches.append(
                                                        RuleMatch(
                                                            proppath,
                                                            message.format(
                                                                prop, k, resourcename
                                                            ),
                                                        )
                                                    )
                            else:
                                message = "Property {0} should be of type List for resource {1}"
                                matches.append(
                                    RuleMatch(
                                        proppath, message.format(prop, resourcename)
                                    )
                                )
                    else:
                        if resourcespec[prop]["Type"] not in ["Map"]:
                            matches.extend(
                                self.propertycheck(
                                    text[prop],
                                    resourcespec[prop]["Type"],
                                    parenttype,
                                    resourcename,
                                    proppath,
                                    False,
                                )
                            )
                elif "PrimitiveType" in resourcespec[prop]:
                    primtype = resourcespec[prop]["PrimitiveType"]
                    matches.extend(
                        self.primitivetypecheck(text[prop], primtype, proppath)
                    )

        return matches

    def match(self, cfn):
        """Check CloudFormation Properties"""
        matches = []
        self.cfn = cfn

        resourcespecs = cfnlint.helpers.RESOURCE_SPECS[cfn.regions[0]]
        self.resourcetypes = resourcespecs["ResourceTypes"]
        self.propertytypes = resourcespecs["PropertyTypes"]
        self.intrinsictypes = resourcespecs["IntrinsicTypes"]
        self.parameternames = self.cfn.get_parameter_names()
        for resourcename, resourcevalue in cfn.get_resources().items():
            if "Properties" in resourcevalue and "Type" in resourcevalue:
                resourcetype = resourcevalue.get("Type", None)
                if (
                    resourcetype.startswith("Custom::")
                    and resourcetype not in self.resourcetypes
                ):
                    resourcetype = "AWS::CloudFormation::CustomResource"
                if resourcetype in self.resourcetypes:
                    path = ["Resources", resourcename, "Properties"]
                    matches.extend(
                        self.propertycheck(
                            resourcevalue.get("Properties", {}),
                            "",
                            resourcetype,
                            resourcename,
                            path,
                            True,
                        )
                    )

        return matches



"""
Helpers for loading resources, managing specs, constants, etc.

Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import datetime
import fnmatch
import gzip
import hashlib
import importlib
import importlib.resources as pkg_resources
import inspect
import json
import logging
import os
import sys
from io import BytesIO
from typing import Dict, List
from urllib.request import Request, urlopen

import regex as re

from cfnlint.data import CloudSpecs

LOGGER = logging.getLogger(__name__)

SPEC_REGIONS = {
    "af-south-1": "https://cfn-resource-specifications-af-south-1-prod.s3.af-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-east-1": "https://cfn-resource-specifications-ap-east-1-prod.s3.ap-east-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-northeast-1": "https://d33vqc0rt9ld30.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-northeast-2": "https://d1ane3fvebulky.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-northeast-3": "https://d2zq80gdmjim8k.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-south-1": "https://d2senuesg1djtx.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-south-2": "https://cfn-resource-specifications-ap-south-2-prod.s3.ap-south-2.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-southeast-1": "https://doigdx0kgq9el.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-southeast-2": "https://d2stg8d246z9di.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ap-southeast-3": "https://cfn-resource-specifications-ap-southeast-3-prod.s3.ap-southeast-3.amazonaws.com/latest/CloudFormationResourceSpecification.json",
    "ap-southeast-4": "https://cfn-resource-specifications-ap-southeast-4-prod.s3.ap-southeast-4.amazonaws.com/latest/CloudFormationResourceSpecification.json",
    "ca-central-1": "https://d2s8ygphhesbe7.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "ca-west-1": "https://cfn-resource-specifications-ca-west-1-prod.s3.ca-west-1.amazonaws.com/latest/CloudFormationResourceSpecification.json",
    "cn-north-1": "https://cfn-resource-specifications-cn-north-1-prod.s3.cn-north-1.amazonaws.com.cn/latest/gzip/CloudFormationResourceSpecification.json",
    "cn-northwest-1": "https://cfn-resource-specifications-cn-northwest-1-prod.s3.cn-northwest-1.amazonaws.com.cn/latest/gzip/CloudFormationResourceSpecification.json",
    "il-central-1": "https://cfn-resource-specifications-il-central-1-prod.s3.il-central-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-central-1": "https://d1mta8qj7i28i2.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-central-2": "https://cfn-resource-specifications-eu-central-2-prod.s3.eu-central-2.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-north-1": "https://diy8iv58sj6ba.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-south-1": "https://cfn-resource-specifications-eu-south-1-prod.s3.eu-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-south-2": "https://cfn-resource-specifications-eu-south-2-prod.s3.eu-south-2.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-west-1": "https://d3teyb21fexa9r.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-west-2": "https://d1742qcu2c1ncx.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "eu-west-3": "https://d2d0mfegowb3wk.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "me-south-1": "https://cfn-resource-specifications-me-south-1-prod.s3.me-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "me-central-1": "https://cfn-resource-specifications-me-central-1-prod.s3.me-central-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
    "sa-east-1": "https://d3c9jyj3w509b0.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "us-east-1": "https://d1uauaxba7bl26.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "us-east-2": "https://dnwj8swjjbsbt.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "us-gov-east-1": "https://s3.us-gov-east-1.amazonaws.com/cfn-resource-specifications-us-gov-east-1-prod/latest/gzip/CloudFormationResourceSpecification.json",
    "us-gov-west-1": "https://s3.us-gov-west-1.amazonaws.com/cfn-resource-specifications-us-gov-west-1-prod/latest/gzip/CloudFormationResourceSpecification.json",
    "us-west-1": "https://d68hl49wbnanq.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
    "us-west-2": "https://d201a2mn26r7lk.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
}
TAG_MAP = "tag:yaml.org,2002:map"
UNCONVERTED_SUFFIXES = ["Ref", "Condition"]
FN_PREFIX = "Fn::"
CONDITION_FUNCTIONS = ["Fn::If"]
REGIONS = list(SPEC_REGIONS.keys())

REGEX_ALPHANUMERIC = re.compile("^[a-zA-Z0-9]*$")
REGEX_CIDR = re.compile(
    r"^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])(\/([0-9]|[1-2][0-9]|3[0-2]))$"
)
REGEX_IPV4 = re.compile(
    r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)){3}$"
)
REGEX_IPV6 = re.compile(
    r"^(((?=.*(::))(?!.*\3.+\3))\3?|[\dA-F]{1,4}:)([\dA-F]{1,4}(\3|:\b)|\2){5}(([\dA-F]{1,4}(\3|:\b|$)|\2){2}|(((2[0-4]|1\d|[1-9])?\d|25[0-5])\.?\b){4})\Z",
    re.I | re.S,
)
REGEX_DYN_REF = re.compile(r"^.*{{resolve:.+}}.*$")
REGEX_DYN_REF_SSM = re.compile(r"^.*{{resolve:ssm:[a-zA-Z0-9_\.\-/]+(:\d+)?}}.*$")
REGEX_DYN_REF_SSM_SECURE = re.compile(
    r"^.*{{resolve:ssm-secure:[a-zA-Z0-9_\.\-/]+(:\d+)?}}.*$"
)

FUNCTIONS = [
    "Fn::Base64",
    "Fn::GetAtt",
    "Fn::GetAZs",
    "Fn::ImportValue",
    "Fn::Join",
    "Fn::Split",
    "Fn::FindInMap",
    "Fn::Select",
    "Ref",
    "Fn::If",
    "Fn::Contains",
    "Fn::Sub",
    "Fn::Cidr",
    "Fn::Length",
    "Fn::ToJsonString",
]

FUNCTIONS_MULTIPLE = ["Fn::GetAZs", "Fn::Split"]

# FindInMap can be singular or multiple.  This needs to be accounted for individually
FUNCTIONS_SINGLE = list(set(FUNCTIONS) - set(FUNCTIONS_MULTIPLE) - set("Fn::FindInMap"))

FUNCTION_IF = "Fn::If"
FUNCTION_AND = "Fn::And"
FUNCTION_OR = "Fn::Or"
FUNCTION_NOT = "Fn::Not"
FUNCTION_EQUALS = "Fn::Equals"
FUNCTION_FOR_EACH = re.compile(r"^Fn::ForEach::[a-zA-Z0-9]+$")

PSEUDOPARAMS = [
    "AWS::AccountId",
    "AWS::NotificationARNs",
    "AWS::NoValue",
    "AWS::Partition",
    "AWS::Region",
    "AWS::StackId",
    "AWS::StackName",
    "AWS::URLSuffix",
]

LIMITS = {
    "Mappings": {"number": 200, "attributes": 200, "name": 255},  # in characters
    "Outputs": {
        "number": 200,
        "name": 255,  # in characters
        "description": 1024,  # in bytes
    },
    "Parameters": {
        "number": 200,
        "name": 255,  # in characters
        "value": 4096,  # in bytes
    },
    "Resources": {"number": 500, "name": 255},  # in characters
    "template": {"body": 1000000, "description": 1024},  # in bytes  # in bytes
    "threshold": 0.9,  # for rules about approaching the other limit values
}

valid_snapshot_types = [
    "AWS::EC2::Volume",
    "AWS::ElastiCache::CacheCluster",
    "AWS::ElastiCache::ReplicationGroup",
    "AWS::Neptune::DBCluster",
    "AWS::RDS::DBCluster",
    "AWS::RDS::DBInstance",
    "AWS::Redshift::Cluster",
]

VALID_PARAMETER_TYPES_SINGLE = [
    "AWS::EC2::AvailabilityZone::Name",
    "AWS::EC2::Image::Id",
    "AWS::EC2::Instance::Id",
    "AWS::EC2::KeyPair::KeyName",
    "AWS::EC2::SecurityGroup::GroupName",
    "AWS::EC2::SecurityGroup::Id",
    "AWS::EC2::Subnet::Id",
    "AWS::EC2::VPC::Id",
    "AWS::EC2::Volume::Id",
    "AWS::Route53::HostedZone::Id",
    "AWS::SSM::Parameter::Name",
    "Number",
    "String",
    "AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>",
    "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
    "AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>",
    "AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>",
    "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>",
    "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>",
    "AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>",
    "AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>",
    "AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>",
    "AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>",
    "AWS::SSM::Parameter::Value<AWS::SSM::Parameter::Name>",
    "AWS::SSM::Parameter::Value<Number>",
    "AWS::SSM::Parameter::Value<String>",
]

VALID_PARAMETER_TYPES_LIST = [
    "CommaDelimitedList",
    "List<AWS::EC2::AvailabilityZone::Name>",
    "List<AWS::EC2::Image::Id>",
    "List<AWS::EC2::Instance::Id>",
    "List<AWS::EC2::SecurityGroup::GroupName>",
    "List<AWS::EC2::SecurityGroup::Id>",
    "List<AWS::EC2::Subnet::Id>",
    "List<AWS::EC2::VPC::Id>",
    "List<AWS::EC2::Volume::Id>",
    "List<AWS::Route53::HostedZone::Id>",
    "List<Number>",
    "List<String>",
    "AWS::SSM::Parameter::Value<CommaDelimitedList>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>",
    "AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>",
    "AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>",
    "AWS::SSM::Parameter::Value<List<Number>>",
    "AWS::SSM::Parameter::Value<List<String>>",
]

VALID_PARAMETER_TYPES = VALID_PARAMETER_TYPES_SINGLE + VALID_PARAMETER_TYPES_LIST


# pylint: disable=missing-class-docstring
class RegexDict(dict):
    def __getitem__(self, item):
        possible_items = {}
        for k, v in self.items():
            if isinstance(v, dict):
                if v.get("Type") == "MODULE":
                    if re.match(k, item):
                        possible_items[k] = v
                else:
                    if k == item:
                        possible_items[k] = v
            elif re.match(k, item):
                possible_items[k] = v
        if not possible_items:
            raise KeyError
        longest_match = sorted(possible_items.keys(), key=len)[-1]
        return possible_items[longest_match]

    def __contains__(self, item):
        for k, v in self.items():
            if isinstance(v, dict):
                if v.get("Type") == "MODULE":
                    if re.match(k, item):
                        return True
                else:
                    if k == item:
                        return True
            elif re.match(k, item):
                return True
        return False

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default


def get_metadata_filename(url):
    """Returns the filename for a metadata file associated with a remote resource"""
    caching_dir = os.path.join(os.path.dirname(__file__), "data", "DownloadsMetadata")
    encoded_url = hashlib.sha256(url.encode()).hexdigest()
    metadata_filename = os.path.join(caching_dir, encoded_url + ".meta.json")

    return metadata_filename


def url_has_newer_version(url):
    """Checks to see if a newer version of the resource at the URL is available
    Always returns true if using Python2.7 due to lack of HEAD request support,
    or if we have no caching information for the local version of the resource
    """
    metadata_filename = get_metadata_filename(url)

    # Load in the cache
    metadata = load_metadata(metadata_filename)

    # Etag is a caching identifier used by S3 and Cloudfront
    if "etag" in metadata:
        cached_etag = metadata["etag"]
    else:
        # If we don't know the etag of the local version, we should force an update
        return True

    # Need to wrap this in a try, as URLLib2 in Python2 doesn't support HEAD requests
    try:
        # Make an initial HEAD request
        req = Request(url, method="HEAD")
        with urlopen(req) as res:
            # If we have an ETag value stored and it matches the returned one,
            # then we already have a copy of the most recent version of the
            # resource, so don't bother fetching it again
            if (
                cached_etag
                and res.info().get("ETag")
                and cached_etag == res.info().get("ETag")
            ):
                LOGGER.debug(
                    "We already have a cached version of url %s with ETag value of %s",
                    url,
                    cached_etag,
                )
                return False

    except NameError:
        # We should force an update
        return True

    # The ETag value of the remote resource does not match the local one, so a newer version is available
    return True


def get_url_content(url, caching=False):
    """Get the contents of a spec file"""

    with urlopen(url) as res:
        if caching and res.info().get("ETag"):
            metadata_filename = get_metadata_filename(url)
            # Load in all existing values
            metadata = load_metadata(metadata_filename)
            metadata["etag"] = res.info().get("ETag")
            metadata["url"] = url  # To make it obvious which url the Tag relates to
            save_metadata(metadata, metadata_filename)

        # Continue to handle the file download normally
        if res.info().get("Content-Encoding") == "gzip":
            buf = BytesIO(res.read())
            f = gzip.GzipFile(fileobj=buf)
            content = f.read().decode("utf-8")
        else:
            content = res.read().decode("utf-8")

    return content


def load_metadata(filename):
    """Get the contents of the download metadata file"""
    metadata = {}
    if os.path.exists(filename):
        with open(filename, "r", encoding="utf-8") as metadata_file:
            metadata = json.load(metadata_file)
    return metadata


def save_metadata(metadata, filename):
    """Save the contents of the download metadata file"""
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.mkdir(dirname)

    with open(filename, "w", encoding="utf-8") as metadata_file:
        json.dump(metadata, metadata_file)


def load_resource(package, filename="us-east-1.json"):
    """Load CloudSpec resources
    :param filename: filename to load
    :return: Json output of the resource laoded
    """
    if sys.version_info >= (3, 9):
        return json.loads(
            pkg_resources.files(package)  # pylint: disable=no-member
            .joinpath(filename)
            .read_text(encoding="utf-8")
        )
    # pylint: disable=W4902
    return json.loads(pkg_resources.read_text(package, filename, encoding="utf-8"))


RESOURCE_SPECS: Dict[str, dict] = {}
REGISTRY_SCHEMAS: List[dict] = []


def merge_spec(source, destination):
    """Recursive merge spec dict"""

    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            merge_spec(value, node)
        else:
            destination[key] = value

    return destination


def set_specs(override_spec_data):
    """Override Resource Specs"""

    excludes = []
    includes = []

    # Extract the exclude list from the override file
    if "ExcludeResourceTypes" in override_spec_data:
        excludes = override_spec_data.pop("ExcludeResourceTypes")
    if "IncludeResourceTypes" in override_spec_data:
        includes = override_spec_data.pop("IncludeResourceTypes")

    for region, spec in RESOURCE_SPECS.items():
        # Merge override spec file into the AWS Resource specification
        if override_spec_data:
            RESOURCE_SPECS[region] = merge_spec(override_spec_data, spec)

        # Grab a list of all resources
        all_resources = list(RESOURCE_SPECS[region]["ResourceTypes"].keys())[:]

        resources = []

        # Remove unsupported resource using includes
        if includes:
            for include in includes:
                regex = re.compile(include.replace("*", "(.*)") + "$")
                matches = [
                    string for string in all_resources if re.match(regex, string)
                ]

                resources.extend(matches)
        else:
            resources = all_resources[:]

        # Remove unsupported resources using the excludes
        if excludes:
            for exclude in excludes:
                regex = re.compile(exclude.replace("*", "(.*)") + "$")
                matches = [string for string in resources if re.match(regex, string)]

                for match in matches:
                    resources.remove(match)

        # Remove unsupported resources
        for resource in all_resources:
            if resource not in resources:
                del RESOURCE_SPECS[region]["ResourceTypes"][resource]


def is_custom_resource(resource_type):
    """Return True if resource_type is a custom resource"""
    return resource_type and (
        resource_type == "AWS::CloudFormation::CustomResource"
        or resource_type.startswith("Custom::")
    )


def bool_compare(first, second):
    """Compare strings to boolean values"""

    if isinstance(first, str):
        first = bool(first.lower() in ["true", "True"])

    if isinstance(second, str):
        second = bool(second.lower() in ["true", "True"])

    return first is second


def initialize_specs():
    """Reload Resource Specs"""

    def load_region(region):
        spec = load_resource(CloudSpecs, filename=f"{region}.json")

        for section, section_values in spec.items():
            if section in ["ResourceTypes", "PropertyTypes", "ValueTypes"]:
                for key, value in section_values.items():
                    if value == "CACHED" and RESOURCE_SPECS["us-east-1"][section].get(
                        key
                    ):
                        spec[section][key] = RESOURCE_SPECS["us-east-1"][section][key]
        return spec

    RESOURCE_SPECS["us-east-1"] = load_region("us-east-1")
    for region in REGIONS:
        if region != "us-east-1":
            RESOURCE_SPECS[region] = load_region(region)


initialize_specs()


def format_json_string(json_string):
    """Format the given JSON string"""

    def converter(o):  # pylint: disable=R1710
        """Help convert date/time into strings"""
        if isinstance(o, datetime.datetime):
            return o.__str__()  # pylint: disable=unnecessary-dunder-call

    return json.dumps(
        json_string, indent=1, sort_keys=True, separators=(",", ": "), default=converter
    )


def create_rules(mod):
    """Create and return an instance of each CloudFormationLintRule subclass
    from the given module."""
    result = []
    for _, clazz in inspect.getmembers(mod, inspect.isclass):
        if (
            clazz.__name__ == "CustomRule"
            and clazz.__module__ == "cfnlint.rules.custom"
        ):
            continue
        method_resolution = inspect.getmro(clazz)
        if [
            clz
            for clz in method_resolution[1:]
            if clz.__module__ in ("cfnlint", "cfnlint.rules")
            and clz.__name__ == "CloudFormationLintRule"
        ]:
            # create and instance of subclasses of CloudFormationLintRule
            obj = clazz()
            result.append(obj)
    return result


loader_details = (
    importlib.machinery.SourceFileLoader,  # pylint: disable=no-member
    importlib.machinery.SOURCE_SUFFIXES,  # pylint: disable=no-member
)


def import_filename(pluginname, root):
    """import_filename imports a module from a file"""
    mod_finder = importlib.machinery.FileFinder(  # pylint: disable=no-member
        root, loader_details
    )

    mod_spec = mod_finder.find_spec(pluginname)
    if mod_spec is not None:
        # for python 2.7 disabling pylint checks
        mod = importlib.util.module_from_spec(mod_spec)  # pylint: disable=no-member
        mod_spec.loader.exec_module(mod)
        return mod

    return None


def load_plugins(directory):
    """Load plugins"""
    result = []

    def onerror(os_error):
        """Raise an error"""
        raise os_error

    for root, _, filenames in os.walk(directory, onerror=onerror):
        for filename in fnmatch.filter(filenames, "[A-Za-z]*.py"):
            mod = import_filename(filename.replace(".py", ""), root)
            if mod is not None:
                result.extend(create_rules(mod))

    return result


def override_specs(override_spec_file):
    """Override specs file"""
    try:
        filename = override_spec_file
        with open(filename, encoding="utf-8") as fp:
            custom_spec_data = json.load(fp)

        set_specs(custom_spec_data)
    except IOError as e:
        if e.errno == 2:
            LOGGER.error("Override spec file not found: %s", filename)
            sys.exit(1)
        elif e.errno == 21:
            LOGGER.error(
                "Override spec file references a directory, not a file: %s", filename
            )
            sys.exit(1)
        elif e.errno == 13:
            LOGGER.error(
                "Permission denied when accessing override spec file: %s", filename
            )
            sys.exit(1)
    except ValueError as err:
        LOGGER.error("Override spec file %s is malformed: %s", filename, err)
        sys.exit(1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from datetime import datetime

from cfnlint.rules import RuleMatch
from cfnlint.rules.resources.lmbd.DeprecatedRuntime import DeprecatedRuntime


class DeprecatedRuntimeEol(DeprecatedRuntime):
    """Check if EOL Lambda Function Runtimes are used"""

    id = "W2531"
    shortdesc = "Check if EOL Lambda Function Runtimes are used"
    description = (
        "Check if an EOL Lambda Runtime is specified and give a warning if used. "
    )
    source_url = (
        "https://docs.aws.amazon.com/lambda/latest/dg/runtime-support-policy.html"
    )
    tags = ["resources", "lambda", "runtime"]

    def check_runtime(self, runtime_value, path):
        """Check if the given runtime is valid"""
        matches = []

        runtime = self.deprecated_runtimes.get(runtime_value)
        if runtime:
            if (
                datetime.strptime(runtime["eol"], "%Y-%m-%d") < self.current_date
                and datetime.strptime(runtime["deprecated"], "%Y-%m-%d")
                > self.current_date
            ):
                message = "EOL runtime ({0}) specified. Runtime is EOL since {1} and updating will be disabled at {2}. Please consider updating to {3}"
                matches.append(
                    RuleMatch(
                        path,
                        message.format(
                            runtime_value,
                            runtime["eol"],
                            runtime["deprecated"],
                            runtime["successor"],
                        ),
                    )
                )
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule
from cfnlint.rules.conditions.common import check_condition_list


class Not(CloudFormationLintRule):
    """Check Not Condition Function Logic"""

    id = "E8005"
    shortdesc = "Check Fn::Not structure for validity"
    description = "Check Fn::Not is a list of one element"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-not"
    tags = ["functions", "not"]

    def match(self, cfn):
        return check_condition_list(cfn, "Fn::Not")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import bool_compare
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class NoEcho(CloudFormationLintRule):
    id = "W4002"
    shortdesc = "Check for NoEcho References"
    description = "Check if there is a NoEcho enabled parameter referenced within a resources Metadata section"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html#parameters-section-structure-properties"
    tags = ["resources", "NoEcho"]

    def _get_no_echo_params(self, cfn):
        """Get no Echo Params"""
        no_echo_params = []
        for parameter_name, parameter_value in cfn.get_parameters().items():
            if isinstance(parameter_value, dict):
                noecho = parameter_value.get("NoEcho", default=False)
                if bool_compare(noecho, True):
                    no_echo_params.append(parameter_name)
        return no_echo_params

    def _check_ref(self, cfn, no_echo_params):
        """Check Refs"""
        matches = []
        refs = cfn.search_deep_keys("Ref")
        for ref in refs:
            if ref[-1] in no_echo_params:
                if len(ref) > 3:
                    if ref[0] == "Resources" and ref[2] == "Metadata":
                        matches.append(
                            RuleMatch(
                                ref,
                                'As the resource "metadata" section contains '
                                + 'reference to a "NoEcho" parameter '
                                + str(ref[-1])
                                + ", CloudFormation will display the parameter value in "
                                + "plaintext",
                            )
                        )

        return matches

    def _check_sub(self, cfn, no_echo_params):
        """Check Subs"""
        matches = []
        subs = cfn.search_deep_keys("Fn::Sub")
        for sub in subs:
            if isinstance(sub[-1], str):
                params = cfn.get_sub_parameters(sub[-1])
                for param in params:
                    if param in no_echo_params:
                        if len(sub) > 2:
                            if sub[0] == "Resources" and sub[2] == "Metadata":
                                matches.append(
                                    RuleMatch(
                                        sub[:-1],
                                        'As the resource "metadata" section contains '
                                        + 'reference to a "NoEcho" parameter '
                                        + str(param)
                                        + ", CloudFormation will display the parameter value in "
                                        + "plaintext",
                                    )
                                )

        return matches

    def match(self, cfn):
        matches = []
        no_echo_params = self._get_no_echo_params(cfn)
        if not no_echo_params:
            return matches
        matches.extend(self._check_ref(cfn, no_echo_params))
        matches.extend(self._check_sub(cfn, no_echo_params))

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import VALID_PARAMETER_TYPES_LIST
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class Equals(CloudFormationLintRule):
    """Check Equals Condition Function Logic"""

    id = "E8003"
    shortdesc = "Check Fn::Equals structure for validity"
    description = "Check Fn::Equals is a list of two elements"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-equals"
    tags = ["functions", "equals"]

    allowed_functions = [
        "Ref",
        "Fn::FindInMap",
        "Fn::Sub",
        "Fn::Join",
        "Fn::Select",
        "Fn::Split",
        "Fn::Length",
        "Fn::ToJsonString",
    ]
    function = "Fn::Equals"

    def _check_equal_values(self, element, path, valid_refs):
        matches = []

        if len(element) == 1:
            for element_key, element_value in element.items():
                if element_key not in self.allowed_functions:
                    message = (
                        self.function + " element must be a supported function ({0})"
                    )
                    matches.append(
                        RuleMatch(
                            path[:] + [element_key],
                            message.format(", ".join(self.allowed_functions)),
                        )
                    )
                elif element_key == "Ref":
                    valid_ref = valid_refs.get(element_value)
                    if valid_ref:
                        if valid_ref.get("From") == "Parameters":
                            if valid_ref.get("Type") in VALID_PARAMETER_TYPES_LIST:
                                message = "Every Fn::Equals object requires a list of 2 string parameters"
                                matches.append(RuleMatch(path, message))
        else:
            message = self.function + " element must be a supported function ({0})"
            matches.append(
                RuleMatch(path, message.format(", ".join(self.allowed_functions)))
            )
        return matches

    def match(self, cfn):
        matches = []
        # Build the list of functions
        trees = cfn.search_deep_keys(self.function)

        valid_refs = cfn.get_valid_refs()
        for tree in trees:
            # Test when in Conditions
            if tree[0] == "Conditions":
                value = tree[-1]
                if not isinstance(value, list):
                    message = self.function + " must be a list of two elements"
                    matches.append(RuleMatch(tree[:-1], message.format()))
                elif len(value) != 2:
                    message = self.function + " must be a list of two elements"
                    matches.append(RuleMatch(tree[:-1], message.format()))
                else:
                    for index, element in enumerate(value):
                        if isinstance(element, dict):
                            matches.extend(
                                self._check_equal_values(
                                    element, tree[:-1] + [index], valid_refs
                                )
                            )
                        elif not isinstance(element, (str, bool, int, float)):
                            message = (
                                self.function
                                + " element must be a String, Boolean, Number, or supported function ({0})"
                            )
                            matches.append(
                                RuleMatch(
                                    tree[:-1] + [index],
                                    message.format(", ".join(self.allowed_functions)),
                                )
                            )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule, RuleMatch


class RuleScheduleExpression(CloudFormationLintRule):
    """Validate AWS Events Schedule expression format"""

    id = "E3027"
    shortdesc = "Validate AWS Event ScheduleExpression format"
    description = "Validate the formation of the AWS::Event ScheduleExpression"
    source_url = "https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/ScheduledEvents.html"
    tags = ["resources", "events"]

    def initialize(self, cfn):
        """Initialize the rule"""
        self.resource_property_types = ["AWS::Events::Rule"]

    def check_rate(self, value, path):
        """Check Rate configuration"""
        matches = []
        # Extract the expression from rate(XXX)
        rate_expression = value[value.find("(") + 1 : value.find(")")]

        if not rate_expression:
            return [RuleMatch(path, "Rate value of ScheduleExpression cannot be empty")]

        # Rate format: rate(Value Unit)
        items = rate_expression.split(" ")

        if len(items) != 2:
            message = "Rate expression must contain 2 elements (Value Unit), rate contains {} elements"
            matches.append(RuleMatch(path, message.format(len(items))))
            return [RuleMatch(path, message.format(len(items)))]

        # Check the Value
        if not items[0].isdigit():
            message = "Rate Value ({}) should be of type Integer."
            extra_args = {
                "actual_type": type(items[0]).__name__,
                "expected_type": int.__name__,
            }
            return [RuleMatch(path, message.format(items[0]), **extra_args)]

        if float(items[0]) <= 0:
            return [
                RuleMatch(path, f"Rate Value {items[0]!r} should be greater than 0.")
            ]

        if float(items[0]) <= 1:
            valid_periods = ["minute", "hour", "day"]
        elif float(items[0]) > 1:
            valid_periods = ["minutes", "hours", "days"]
        # Check the Unit
        if items[1] not in valid_periods:
            return [
                RuleMatch(
                    path, f"Rate Unit {items[1]!r} should be one of {valid_periods!r}."
                )
            ]

        return []

    def check_cron(self, value, path):
        """Check Cron configuration"""
        matches = []
        # Extract the expression from cron(XXX)
        cron_expression = value[value.find("(") + 1 : value.find(")")]

        if not cron_expression:
            matches.append(
                RuleMatch(path, "Cron value of ScheduleExpression cannot be empty")
            )
        else:
            # Rate format: cron(Minutes Hours Day-of-month Month Day-of-week Year)
            items = cron_expression.split(" ")

            if len(items) != 6:
                message = "Cron expression must contain 6 elements (Minutes Hours Day-of-month Month Day-of-week Year), cron contains {} elements"
                matches.append(RuleMatch(path, message.format(len(items))))
                return matches

            _, _, day_of_month, _, day_of_week, _ = cron_expression.split(" ")
            if day_of_month != "?" and day_of_week != "?":
                matches.append(
                    RuleMatch(
                        path,
                        "Don't specify the Day-of-month and Day-of-week fields in the same cron expression",
                    )
                )

        return matches

    def check_value(self, value, path):
        """Count ScheduledExpression value"""
        matches = []

        # Value is either "cron()" or "rate()"
        if value.startswith("rate(") and value.endswith(")"):
            matches.extend(self.check_rate(value, path))
        elif value.startswith("cron(") and value.endswith(")"):
            matches.extend(self.check_cron(value, path))
        else:
            message = "Invalid ScheduledExpression specified ({}). Value has to be either cron() or rate()"
            matches.append(RuleMatch(path, message.format(value)))

        return matches

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""
        matches = []

        matches.extend(
            cfn.check_value(
                obj=properties,
                key="ScheduleExpression",
                path=path[:],
                check_value=self.check_value,
            )
        )

        return matches
